use poise::CreateReply;
use reqwest::Client;
use serenity::all::ButtonStyle;
use serenity::all::{
    ChannelId, CreateActionRow, CreateAllowedMentions, CreateButton, CreateMessage, ReactionType,
    User,
};
use sqlx::{Pool, Sqlite};

use crate::commands::command_helper::cooldown;
use crate::data::account_links::{Link, LinkId, Uuid};
use crate::error::Error;
use crate::error::Error::Other;
use crate::Caches;
use crate::Context;

#[poise::command(
    slash_command,
    subcommands("add", "list"),
    install_context = "User|Guild",
    interaction_context = "Guild|BotDm|PrivateChannel"
)]
pub(crate) async fn account(_ctx: Context<'_>) -> Result<(), Error> {
    // root of slash-commands is not invokable.
    unreachable!()
}

#[poise::command(slash_command, ephemeral = "false")]
/// Verify a Minecraft account on the Zombies MultiPlayer Discord.
pub(crate) async fn add(
    ctx: Context<'_>,
    #[description = "Minecraft username"]
    #[min_length = 2]
    #[max_length = 16]
    ign: String,
    #[description = "Discord User"] user: Option<User>,
    #[description = "admin-only"] force: Option<bool>,
) -> Result<(), Error> {
    ctx.defer().await?;
    let force: bool =
        force.unwrap_or(false) && ctx.framework().options.owners.contains(&ctx.author().id) && {
            let _ = user.as_ref().ok_or(Other(
                "Warning: attempted to run forced account add without specifying a target Discord \
                 account."
                    .to_string(),
            ))?;
            true
        };
    let user: User = user.unwrap_or(ctx.author().clone());
    let uuid: Uuid = Uuid::for_ign(
        ign.as_str(),
        &ctx.data().clients.general,
        &ctx.data().caches,
    )
    .await?;
    match force
        || uuid
            .has_discord_user(&user, &ctx.data().clients.hypixel_api_client)
            .await?
    {
        true => {
            let pool = &ctx.data().sqlite_pool;
            let status: &str = match LinkId::try_from_minecraft(pool, uuid.as_str()).await {
                Err(_) => match LinkId::try_from_discord(pool, user.id.get()).await {
                    Err(_) => {
                        let id = LinkId::new(pool).await?;
                        sqlx::query(
                            format!(
                                "INSERT INTO discord_links VALUES ({id}, {});",
                                user.id.get()
                            )
                            .as_str(),
                        )
                        .execute(pool)
                        .await?;
                        sqlx::query(
                            format!(
                                "INSERT INTO minecraft_links VALUES ({id}, \"{}\");",
                                uuid.get()
                            )
                            .as_str(),
                        )
                        .execute(pool)
                        .await?;
                        "Linked your Discord and Minecraft account."
                    }
                    Ok(dc_id) => {
                        sqlx::query(
                            format!(
                                "INSERT INTO minecraft_links VALUES ({dc_id}, \"{}\");",
                                uuid.get()
                            )
                            .as_str(),
                        )
                        .execute(pool)
                        .await?;
                        "Your Discord account has previously had an account linked. Added the new \
                         link."
                    }
                },
                Ok(mc_id) => match LinkId::try_from_discord(pool, user.id.get()).await {
                    Err(_) => {
                        sqlx::query(
                            format!(
                                "INSERT INTO discord_links VALUES ({mc_id}, {});",
                                user.id.get()
                            )
                            .as_str(),
                        )
                        .execute(pool)
                        .await?;
                        "Your Minecraft account has previously had an account linked. Added the \
                         new link."
                    }
                    Ok(dc_id) => {
                        sqlx::query(
                            format!(
                                "UPDATE minecraft_links SET link_id = {mc_id} WHERE link_id = {dc_id};",
                            )
                            .as_str(),
                        )
                        .execute(pool)
                        .await?;
                        sqlx::query(
                            format!(
                                "UPDATE discord_links SET link_id = {mc_id} WHERE link_id = {dc_id};",
                            )
                            .as_str(),
                        )
                        .execute(pool)
                        .await?;
                        "Both your Discord and Minecraft account had linked accounts. Merged all \
                         account links."
                    }
                },
            };
            let s = format!(
                "Verification request for <@{}> with IGN `{}`",
                user.id.get(),
                ign
            );
            ChannelId::new(1257776992497959075)
                .send_message(
                    ctx,
                    CreateMessage::new()
                        .content(s)
                        .allowed_mentions(
                            CreateAllowedMentions::new().empty_roles().all_users(true),
                        )
                        .components(vec![CreateActionRow::Buttons(vec![
                            CreateButton::new("accept_verification")
                                .emoji(ReactionType::from('✅'))
                                .style(ButtonStyle::Secondary),
                            CreateButton::new("deny_verification")
                                .emoji(ReactionType::from('❌'))
                                .style(ButtonStyle::Secondary),
                            CreateButton::new("list_accounts")
                                .emoji(ReactionType::from('📜'))
                                .style(ButtonStyle::Primary),
                        ])]),
                )
                .await?;
            ctx.send(CreateReply::default().content(status)).await?;
            Ok(())
        }
        false => Err(Error::Other(format!(
            "The Discord account linked on Hypixel does not match the specified discord \
             account.\nPlease set your linked Discord account on Hypixel to `{}`.",
            user.name
        ))),
    }
}

#[poise::command(
    slash_command,
    ephemeral = "true",
    context_menu_command = "Account list"
)]
/// List a users linked minecraft Accounts.
pub(crate) async fn list(ctx: Context<'_>, user: User) -> Result<(), Error> {
    ctx.defer().await?;
    cooldown(&ctx, 600, 300)?;
    let pool: &Pool<Sqlite> = &ctx.data().sqlite_pool;
    let s: String =
        list_string(pool, &user, &ctx.data().caches, &ctx.data().clients.general).await?;
    ctx.send(
        CreateReply::default()
            .content(s)
            .allowed_mentions(CreateAllowedMentions::new().empty_roles().empty_users()),
    )
    .await?;
    Ok(())
}

pub(crate) async fn list_string(
    pool: &Pool<Sqlite>,
    user: &User,
    c: &Caches,
    cli: &Client,
) -> Result<String, Error> {
    let link = Link::try_from_discord(user, pool).await?;
    let mut discord_list = String::from("### Discord:");
    for dc in link.discord_ids() {
        discord_list.push_str(format!("\n- <@{dc}>").as_str());
    }
    let mut minecraft_list = String::from("### Minecraft:");
    for mc in link.minecraft_accounts() {
        minecraft_list.push_str(format!("\n- `{}`", mc.ign(c, cli).await?).as_str());
    }
    Ok(format!(
        "## Account list for member #{}:\n{}\n{}",
        link.link_id(),
        discord_list.as_str(),
        minecraft_list.as_str()
    ))
}

#[poise::command(slash_command)]
pub(crate) async fn remove(_ctx: Context<'_>) -> Result<(), Error> {
    unimplemented!();
}
